#ifndef EDITORNODE_H
#define EDITORNODE_H

#include <iostream>
#include <memory>
#include <functional>
#include <deque>
#include <algorithm>

template <class T>
class TreeNode;

enum class TreeWalkStrategy {
    depthFirst,
    breadthFirst
};
template <class T>
class TreeNode : public std::enable_shared_from_this<TreeNode<T>>
{
public:
    explicit TreeNode();
    explicit TreeNode(std::shared_ptr<T> data);
    virtual ~TreeNode();

    std::shared_ptr<TreeNode<T>> copy();  // copy only this node, without children!
    std::shared_ptr<TreeNode<T>> rcopy();  // recursive copy

    const std::shared_ptr<TreeNode<T>> nextSibling();

    const std::shared_ptr<TreeNode<T>> next(TreeWalkStrategy walkStrategy = TreeWalkStrategy::depthFirst);
    const std::shared_ptr<TreeNode<T>> getRoot();
    void move(const std::shared_ptr<TreeNode<T> > &to); //as child
    void moveChildren(const std::shared_ptr<TreeNode<T> > &to); //all children but not the 'executing' element

    void addChild(const std::shared_ptr<TreeNode<T> > &child);
    void addChildren(const std::shared_ptr<TreeNode<T> > &nodeWithChildren);
    void removeChild(const std::shared_ptr<TreeNode<T> > &child);
    void remove();

    void addChildObject(const std::shared_ptr<T> &obj);

    uint getDistanceFromRoot();

    std::shared_ptr<TreeNode<T>> getParent() const;
    void setParent(const std::shared_ptr<TreeNode<T>> &value);

    std::shared_ptr<T> getData() const;
    void setData(const std::shared_ptr<T> &value);

    std::shared_ptr<std::deque<std::shared_ptr<TreeNode<T> > > > getChildren() const;

    void setDataChangedSignal(std::function<void(std::shared_ptr<TreeNode<T>>)> f)
    {
        nodeDataChanged = f;
    }
    void setTreeChangedSignal(std::function<void(std::shared_ptr<TreeNode<T>>)> f)
    {
        treeChanged = f;
    }

    std::function<void(std::shared_ptr<TreeNode<T>>)> getTreeChangedSignal(){ return treeChanged; }
    std::function<void(std::shared_ptr<TreeNode<T>>)> getDataChangedSignal(){ return nodeDataChanged; }

private:
    void fireNodeDataChanged();
    void fireTreeChanged();
    //signal callbacks only set in the root node
    std::function<void(std::shared_ptr<TreeNode<T>>)> nodeDataChanged;
    std::function<void(std::shared_ptr<TreeNode<T>>)> treeChanged;

    std::shared_ptr<TreeNode<T>> nextOnSameLevel();
    std::shared_ptr<TreeNode<T>> parent;
    std::shared_ptr<T> data;
    std::shared_ptr<std::deque<std::shared_ptr<TreeNode<T>>>> children;
};

template <class T>
TreeNode<T>::TreeNode()
{
    this->data = nullptr;
    this->nodeDataChanged = nullptr;
    this->treeChanged = nullptr;
    this->children = std::make_shared<std::deque<std::shared_ptr<TreeNode<T>>>>();
}

template <class T>
TreeNode<T>::TreeNode(std::shared_ptr<T> data) : TreeNode()
{
    this->data = data;
}

template <class T>
TreeNode<T>::~TreeNode()
{

}

template<class T>
std::shared_ptr<TreeNode<T> > TreeNode<T>::copy()
{
    return std::make_shared<TreeNode<T>>(std::make_shared<T>(*this->getData()));
}

template<class T>
std::shared_ptr<TreeNode<T> > TreeNode<T>::rcopy()
{
    auto n = std::make_shared<TreeNode<T>>(std::make_shared<T>(*this->getData()));
    for(auto c : *children)
        n->addChild(c->rcopy());
    return n;
}

template <class T>
const std::shared_ptr<TreeNode<T>> TreeNode<T>::nextSibling()
{
    if (parent != nullptr) {
        auto l = *(parent->getChildren());
        auto it = std::find(l.begin(), l.end(), this->shared_from_this());
        if (it != l.end() && it+1 != l.end())
            return *(++it); //next sibling...
        else
            return nullptr;
    }
    return nullptr;

}

template <class T>
const std::shared_ptr<TreeNode<T>> TreeNode<T>::next(TreeWalkStrategy walkStrategy)
{
    if (walkStrategy == TreeWalkStrategy::depthFirst)
        if (getChildren()->size() > 0)
            return getChildren()->front();
        else {
            auto node = this->shared_from_this();
            while (node->nextSibling() == nullptr && node->getParent() != nullptr)
                node = node->getParent();
            return node->nextSibling();
        }
    else if (walkStrategy == TreeWalkStrategy::breadthFirst)
        if (parent != nullptr) {
            auto nextOnLvl = nextOnSameLevel();
            if (nextOnLvl == nullptr){
                for (auto n : *(parent->getChildren()))
                    if (n->getChildren()->size() > 0)
                        return n->getChildren()->front();
            }
            else return nextOnLvl;
        }
    return nullptr;
}

template <class T>
const std::shared_ptr<TreeNode<T>> TreeNode<T>::getRoot()
{
    auto node = this->shared_from_this();
    while (node->getParent() != nullptr)
        node = node->getParent();
    return node;
}

template <class T>
void TreeNode<T>::move(const std::shared_ptr<TreeNode<T>> &to)
{
    this->getParent()->removeChild(this->shared_from_this());
    to->getChildren()->push_back(this->shared_from_this());
    fireTreeChanged();
}

template <class T>
void TreeNode<T>::moveChildren(const std::shared_ptr<TreeNode<T>> &to)
{
    for (auto child : getChildren().get())
        child->move(to);
    fireTreeChanged();
}

template <class T>
void TreeNode<T>::addChild(const std::shared_ptr<TreeNode<T>>& child)
{
    this->getChildren()->push_back(child);
    child->setParent(this->shared_from_this());
    fireTreeChanged();
}


template <class T>
void TreeNode<T>::addChildren(const std::shared_ptr<TreeNode<T>>& nodeWithChildren)
{
    for(std::shared_ptr<TreeNode<T>> child : nodeWithChildren->getChildren().get())
    {
        this->getChildren()->push_back(child);
        child->setParent(this->shared_from_this());
    }
    fireTreeChanged();
}

template <class T>
void TreeNode<T>::removeChild(const std::shared_ptr<TreeNode<T>>& child)
{
    auto it = children->begin();
    while ( *it != child && it != children->end())
        ++it;
    children->erase(it);
    fireTreeChanged();
}

template <class T>
void TreeNode<T>::remove()
{

    if (children->size() > 0) {
        for (auto node : *children) {
            node->remove();
        }
    }

    if (parent != nullptr) {
        auto n = this->shared_from_this();
        auto it = getParent()->getChildren()->begin();
        while ( *it != n && it != getParent()->getChildren()->end())
            ++it;
        fireTreeChanged();

        parent->getChildren()->erase(it);

    }
}

template<class T>
void TreeNode<T>::addChildObject(const std::shared_ptr<T> &obj)
{
    auto node = std::make_shared<TreeNode<T>>(obj);
    this->addChild(node);
}

template <class T>
std::shared_ptr<TreeNode<T>> TreeNode<T>::getParent() const
{
    return parent;
}

template <class T>
void TreeNode<T>::setParent(const std::shared_ptr<TreeNode<T>> &value)
{
    parent = value;
}

template <class T>
std::shared_ptr<T> TreeNode<T>::getData() const
{
    return data;
}

template <class T>
void TreeNode<T>::setData(const std::shared_ptr<T> &value)
{
    data = value;
    fireNodeDataChanged();
}

template <class T>
std::shared_ptr<std::deque<std::shared_ptr<TreeNode<T>>>> TreeNode<T>::getChildren() const
{
    return children;
}

template <class T>
void TreeNode<T>::fireNodeDataChanged()
{
    auto signal = this->getRoot()->getDataChangedSignal();
    if(signal != nullptr)
        signal(this->getParent());
}

template <class T>
void TreeNode<T>::fireTreeChanged()
{

    auto signal = this->getRoot()->getTreeChangedSignal();

    if(signal != nullptr) {

        signal(this->getParent());

    }
}

template <class T>
std::shared_ptr<TreeNode<T>> TreeNode<T>::nextOnSameLevel()
{
    if (nextSibling() != nullptr)
        return nextSibling();
    uint lvl = 1;
    auto node = this->getParent();
    auto pnode = this->shared_from_this();
    while (lvl > 0 && node != nullptr) {
        if (node != pnode && node != pnode->getParent() && node->getChildren()->size() > 0) {
            node = node->getChildren()->front();
            --lvl;
        } else {
            pnode = node;
            if (node->nextSibling() != nullptr)
                node = node->nextSibling();
            else {
                if (node->getParent() != nullptr) {
                    node = node->getParent();
                    ++lvl;
                }
                else return nullptr;
            }
        }
    }
    if (lvl != 0) return nullptr;
    return node;
}

template <class T>
uint TreeNode<T>::getDistanceFromRoot()
{
    std::shared_ptr<TreeNode<T>> node = this->shared_from_this();
    uint lvl = 0;
    while(node->getParent() != nullptr)
    {
        ++lvl;
        node = node->getParent();
    }
    return lvl;
}


#endif // EDITORNODE_H
